#ifndef EXISTING_SYSTEM_MONITOR_INSTRUMENTED_NODE_HPP
#define EXISTING_SYSTEM_MONITOR_INSTRUMENTED_NODE_HPP

#include <string>
#include <memory>
#include <functional>
#include <unordered_map>
#include <chrono>

#include "rclcpp/rclcpp.hpp"
#include "simple_timer/rt-sched.hpp"

namespace existing_system_monitor {

// Base class for instrumented callbacks
class InstrumentedCallback {
public:
    InstrumentedCallback(std::string const &node_name, std::string const &callback_name, int chain_id) 
            : node_name_(node_name)
            , callback_name_(callback_name)
            , chain_id_(chain_id) {
        logger_ = create_logger();
    }
    virtual ~InstrumentedCallback() = default;
    
    template<typename... Args>
    void before_callback(Args&&... args);
    
    template<typename... Args>
    void after_callback(Args&&... args);

    friend class MetricsCollector;

protected:
    node_time_logger logger_;

private:
    std::string node_name_;
    std::string callback_name_;
    int chain_id_;
    std::chrono::high_resolution_clock::time_point start_time_;
};

// Wrapper for a subscription callback
template<typename MessageT>
class InstrumentedSubscriptionCallback : public InstrumentedCallback {
public:
    using CallbackT = std::function<void(typename MessageT::SharedPtr)>;
    
    InstrumentedSubscriptionCallback(
        const std::string& node_name,
        const std::string& topic_name,
        int chain_id,
        CallbackT original_callback);
    
    void callback(typename MessageT::SharedPtr msg) {
        before_callback();
        original_callback_(msg);
        after_callback();
    }

private:
    CallbackT original_callback_;
};

// Wrapper for a timer callback
class InstrumentedTimerCallback : public InstrumentedCallback {
public:
    using CallbackT = std::function<void()>;
    
    InstrumentedTimerCallback(
        const std::string& node_name,
        const std::string& timer_name,
        int chain_id,
        CallbackT original_callback);
    
    void callback();

private:
    CallbackT original_callback_;
};

// Factory class for creating instrumented nodes
class NodeInstrumenter {
public:
    static std::shared_ptr<rclcpp::Node> instrument_node(
        std::shared_ptr<rclcpp::Node> original_node,
        int chain_id = 0);
    
    template<typename MessageT>
    static typename rclcpp::Subscription<MessageT>::SharedPtr instrument_subscription(
        std::shared_ptr<rclcpp::Node> node,
        typename rclcpp::Subscription<MessageT>::SharedPtr original_subscription,
        int chain_id = 0);
    
    static rclcpp::TimerBase::SharedPtr instrument_timer(
        std::shared_ptr<rclcpp::Node> node,
        rclcpp::TimerBase::SharedPtr original_timer,
        int chain_id = 0);
};

// Class to hold metrics collection and analysis
class MetricsCollector {
public:
    struct Config {
        std::string executor_type = "default";
    };

    static MetricsCollector& get_instance();
    
    void register_callback(std::shared_ptr<InstrumentedCallback> callback);
    void write_logs_to_file(const std::string& file_name, const std::string& experiment_name);
    void set_config(Config config) { config_ = config; }

private:
    MetricsCollector() {
        logger_ = create_logger();
    }
    std::unordered_map<std::string, std::shared_ptr<InstrumentedCallback>> callbacks_;
    node_time_logger logger_;
    Config config_;
    std::string buildLogs(std::vector<node_time_logger> node_logs = {});
};

} // namespace existing_system_monitor

#endif // EXISTING_SYSTEM_MONITOR_INSTRUMENTED_NODE_HPP
