#ifndef EXISTING_SYSTEM_MONITOR_NODE_ADAPTER_HPP
#define EXISTING_SYSTEM_MONITOR_NODE_ADAPTER_HPP

#include "rclcpp/rclcpp.hpp"

#include "existing_system_monitor/instrumented_node.hpp"


namespace existing_system_monitor {

/**
 * Adapter class that wraps an existing ROS2 node with instrumentation
 * Usage: instead of inheriting from rclcpp::Node, inherit from InstrumentedNode
 */
template <typename NodeT>
class InstrumentedNode : public NodeT {
public:
    struct NodeIdentifier {
        int chain_id;
        std::string node_name;
    };

    template <typename... Args>
    InstrumentedNode(NodeIdentifier id, Args&&... args) 
        : NodeT(std::forward<Args>(args)...), chain_id_(id.chain_id), node_name_(id.node_name) {
    }

    // Override subscription creation
    template <typename MessageT, typename CallbackT>
    typename rclcpp::Subscription<MessageT>::SharedPtr create_subscription(
        const std::string& topic_name,
        const rclcpp::QoS& qos,
        CallbackT&& callback,
        const rclcpp::SubscriptionOptions& options = rclcpp::SubscriptionOptions())
    {
        // Create a wrapper for the callback
        auto instrumented_callback =
            std::make_shared<InstrumentedSubscriptionCallback<MessageT>>(
                this->get_name(),
                node_name_ + "-" + topic_name,
                chain_id_,
                std::forward<CallbackT>(callback)
        );

        // Register the callback with the MetricsCollector
        MetricsCollector::get_instance().register_callback(instrumented_callback);
        
        // Create the subscription with the instrumented callback
        auto subscription =
            NodeT::template create_subscription<MessageT>(
                topic_name,
                qos,
                [instrumented_callback](typename MessageT::SharedPtr msg) {
                    instrumented_callback->callback(msg);
                },
                options
        );
        
        return subscription;
    }

    // Override timer creation
    template <typename DurationT, typename CallbackT>
    rclcpp::TimerBase::SharedPtr create_wall_timer(
        DurationT period,
        CallbackT&& callback,
        rclcpp::CallbackGroup::SharedPtr group = nullptr)
    {        
        // Create a wrapper for the callback
        auto instrumented_callback =
            std::make_shared<InstrumentedTimerCallback>(
                this->get_name(),
                node_name_ + "-timer",
                chain_id_,
                std::forward<CallbackT>(callback)
        );

        // Register the callback with the MetricsCollector
        MetricsCollector::get_instance().register_callback(instrumented_callback);
        
        // Create the timer with the instrumented callback
        auto timer =
            NodeT::create_wall_timer(
                period,
                [instrumented_callback]() {
                    instrumented_callback->callback();
                },
                group
        );
        
        return timer;
    }

private:
    int chain_id_;
    std::string node_name_;
};

} // namespace existing_system_monitor

#endif // EXISTING_SYSTEM_MONITOR_NODE_ADAPTER_HPP
