#include <sstream>
#include <fstream>
#include <filesystem>

#include "std_msgs/msg/string.hpp"

#include "existing_system_monitor/instrumented_node.hpp"

namespace existing_system_monitor {


template<typename... Args>
void InstrumentedCallback::before_callback(Args&&... args) {
    // Record start time
    start_time_ = std::chrono::high_resolution_clock::now();
    
    // Log start of callback execution
    std::ostringstream ss;
    ss << "{\"operation\": \"start_work\", \"chain\": " << chain_id_
       << ", \"node\": \"" << node_name_ << "/" << callback_name_ << "\", \"count\": 0}";
    log_entry(logger_, ss.str());
}

template<typename... Args>
void InstrumentedCallback::after_callback(Args&&... args) {
    // Calculate execution time
    auto end_time = std::chrono::high_resolution_clock::now();
    auto duration = std::chrono::duration_cast<std::chrono::microseconds>(
        end_time - start_time_).count();
    
    // Log end of callback execution
    std::ostringstream ss;
    ss << "{\"operation\": \"end_work\", \"chain\": " << chain_id_
       << ", \"node\": \"" << node_name_ << "/" << callback_name_ 
       << "\", \"count\": 0, \"duration_us\": " << duration << "}";
    log_entry(logger_, ss.str());
}

template<typename MessageT>
InstrumentedSubscriptionCallback<MessageT>::InstrumentedSubscriptionCallback(
    const std::string& node_name,
    const std::string& topic_name,
    int chain_id,
    CallbackT original_callback)
    : InstrumentedCallback(node_name, topic_name, chain_id),
      original_callback_(original_callback) {
}

InstrumentedTimerCallback::InstrumentedTimerCallback(
    const std::string& node_name,
    const std::string& timer_name,
    int chain_id,
    CallbackT original_callback)
    : InstrumentedCallback(node_name, timer_name, chain_id),
      original_callback_(original_callback) {
}

void InstrumentedTimerCallback::callback() {
    before_callback();
    original_callback_();
    after_callback();
}

std::shared_ptr<rclcpp::Node> NodeInstrumenter::instrument_node(
    std::shared_ptr<rclcpp::Node> original_node,
    int chain_id) {
    // For now, we just return the original node
    // In a more advanced implementation, we might create a proxy node
    return original_node;
}

template<typename MessageT>
typename rclcpp::Subscription<MessageT>::SharedPtr NodeInstrumenter::instrument_subscription(
    std::shared_ptr<rclcpp::Node> node,
    typename rclcpp::Subscription<MessageT>::SharedPtr original_subscription,
    int chain_id) {
    // To be implemented
    return original_subscription;
}

rclcpp::TimerBase::SharedPtr NodeInstrumenter::instrument_timer(
    std::shared_ptr<rclcpp::Node> node,
    rclcpp::TimerBase::SharedPtr original_timer,
    int chain_id) {
    // To be implemented
    return original_timer;
}

MetricsCollector& MetricsCollector::get_instance() {
    static MetricsCollector instance;
    return instance;
}

void MetricsCollector::register_callback(std::shared_ptr<InstrumentedCallback> callback) {
    // Create a unique key for the callback using its node name and callback name
    std::string key = std::to_string(callback->chain_id_) + "_" + 
                     callback->node_name_ + "_" + 
                     callback->callback_name_;
    callbacks_[key] = callback;
}

std::string MetricsCollector::buildLogs(std::vector<node_time_logger> node_logs) {
    std::stringstream output_file;
    
    output_file << "[" << std::endl;
    
    // Add logs from MetricsCollector's logger
    for (auto const& log : *(logger_.recorded_times)) {
        output_file << "{\"entry\": " << log.first 
                   << ", \"time\": " << log.second << "}," << std::endl;
    }

    // Add logs from registered callbacks
    for (auto const& [name, callback] : callbacks_) {
        for (auto const& log : *(callback->logger_.recorded_times)) {
            output_file << "{\"entry\": " << log.first 
                       << ", \"time\": " << log.second << "}," << std::endl;
        }
    }
    
    // Add logs from additional node loggers passed in
    for (auto const& logger : node_logs) {
        for (auto const& log : *(logger.recorded_times)) {
            output_file << "{\"entry\": " << log.first 
                       << ", \"time\": " << log.second << "}," << std::endl;
        }
    }
    
    // Remove the last comma if there are entries
    if (!logger_.recorded_times->empty() || !callbacks_.empty() || !node_logs.empty()) {
        output_file.seekp(-2, std::ios_base::end);
    }
    output_file << "]" << std::endl;
    
    return output_file.str();
}

void MetricsCollector::write_logs_to_file(const std::string& file_name, const std::string& experiment_name) {
    std::filesystem::path results_dir = "results";
    std::filesystem::path exp_dir = results_dir / experiment_name;

    // Create directories if they don't exist
    try {
        std::filesystem::create_directories(exp_dir);
    } catch (std::filesystem::filesystem_error const &e) {
        std::cout << "Could not create results directory: " << e.what() << std::endl;
        return;
    }

    std::filesystem::path output_path = exp_dir / (file_name + "_" + config_.executor_type + ".json");
    
    std::ofstream output_file(output_path);
    output_file << buildLogs();
    output_file.close();

    std::cout << "results written to " << output_path.string() << std::endl;
}

template class InstrumentedSubscriptionCallback<std_msgs::msg::String>;

} // namespace existing_system_monitor
