#include <chrono>
#include <memory>
#include <string>

#include "rclcpp/rclcpp.hpp"
#include "std_msgs/msg/string.hpp"

#include "casestudy_tools/primes_workload.hpp"
#include "existing_system_monitor/node_adapter.hpp"
#include "priority_executor/priority_executor.hpp"
#include "priority_executor/priority_memory_strategy.hpp"

using namespace std::chrono_literals;

// the topic
constexpr char const* TOPIC_NAME = "wildfire_talk";


// Modified Node - Inherit from InstrumentedNode instead of rclcpp::Node
class CustomTalkerNode : public existing_system_monitor::InstrumentedNode<rclcpp::Node> {
public:
    // Pass chain_id to InstrumentedNode constructor
    CustomTalkerNode(int chain_id, std::string const &node_name) : InstrumentedNode({chain_id, node_name}, node_name) {
        // The rest of the code stays the same!
        timer_ = create_wall_timer(
            std::chrono::milliseconds(100), 
            std::bind(&CustomTalkerNode::timer_callback, this));

        publisher_ = this->create_publisher<std_msgs::msg::String>(TOPIC_NAME, 10);
        timer_ = this->create_wall_timer(
            500ms, std::bind(&CustomTalkerNode::timer_callback, this));
    }

    rclcpp::TimerBase::SharedPtr timer_;
    
private:
    void timer_callback() {
        auto message = std_msgs::msg::String();
        message.data = "Hello, world! " + std::to_string(count_++);
        RCLCPP_INFO(this->get_logger(), "Publishing: '%s'", message.data.c_str());
        publisher_->publish(message);
    }

    rclcpp::Publisher<std_msgs::msg::String>::SharedPtr publisher_;
    int count_;
};

// Modified Node - Inherit from InstrumentedNode instead of rclcpp::Node
class CustomListenerNode : public existing_system_monitor::InstrumentedNode<rclcpp::Node> {
public:
    // Pass chain_id to InstrumentedNode constructor
    CustomListenerNode(int chain_id, const std::string& node_name) : InstrumentedNode({chain_id, node_name}, node_name) {
        sub_ = this->create_subscription<std_msgs::msg::String>(
            TOPIC_NAME, 10, std::bind(&CustomListenerNode::topic_callback, this, std::placeholders::_1));
    }

    rclcpp::Subscription<std_msgs::msg::String>::SharedPtr sub_;
    
private:
    void topic_callback(const std_msgs::msg::String::SharedPtr msg) const {
        RCLCPP_INFO(this->get_logger(), "I heard: '%s'", msg->data.c_str());
    }
};


int main(int argc, char* argv[]) {
    // calibrate the dummy load for the current system
    calibrate_dummy_load();

    existing_system_monitor::MetricsCollector::Config config;
    config.executor_type = "edf";
    auto& collector = existing_system_monitor::MetricsCollector::get_instance();
    collector.set_config(config);

    rclcpp::init(argc, argv);

    // Create nodes for the experiment
    /*
    auto talker = std::make_shared<existing_system_monitor::InstrumentedNode<CustomTalkerNode>>(0, "CustomTalkerNode");
    auto listener1 = std::make_shared<existing_system_monitor::InstrumentedNode<CustomListenerNode>>(0, "CustomListenerNode");
    auto listener2 = std::make_shared<existing_system_monitor::InstrumentedNode<CustomListenerNode>>(0, "CustomListenerNode");
    */
    auto talker = std::make_shared<CustomTalkerNode>(0, "CustomTalkerNode");
    auto listener1 = std::make_shared<CustomListenerNode>(0, "CustomListenerNode");
    auto listener2 = std::make_shared<CustomListenerNode>(0, "CustomListenerNode");

    rclcpp::ExecutorOptions options;

    auto strategy = std::make_shared<priority_executor::PriorityMemoryStrategy<>>();
    options.memory_strategy = strategy;
    auto executor = new priority_executor::TimedExecutor(options);

    // must be set to post_execute can set new deadlines
    executor->prio_memory_strategy_ = strategy;

    // the new funcitons in PriorityMemoryStrategy accept the handle of the
    // timer/subscription as the first argument
    strategy->set_executable_deadline(talker->timer_->get_timer_handle(), 1000, priority_executor::ExecutableType::TIMER, 0);
    // you _must_ set the timer_handle for each chain
    strategy->get_priority_settings(talker->timer_->get_timer_handle())->timer_handle = talker->timer_;
    // you _must_ mark the first executable in the chain
    strategy->set_first_in_chain(talker->timer_->get_timer_handle());
    // set the same period and chain_id for each callback in the chain
    strategy->set_executable_deadline(listener1->sub_->get_subscription_handle(), 1000, priority_executor::ExecutableType::SUBSCRIPTION, 0);
    strategy->set_executable_deadline(listener2->sub_->get_subscription_handle(), 1000, priority_executor::ExecutableType::SUBSCRIPTION, 0);
    // you _must_ mark the last executable in the chain (used to keep track of different instances of the same chain)
    strategy->set_last_in_chain(listener2->sub_->get_subscription_handle());
    // add all the nodes to the executor
    executor->add_node(talker);
    executor->add_node(listener1);
    executor->add_node(listener2);

    // if the executor behaves unexpectedly, you can print the priority settings to make sure they are correct
    std::cout << *strategy->get_priority_settings(talker->timer_->get_timer_handle()) << std::endl;
    std::cout << *strategy->get_priority_settings(listener1->sub_->get_subscription_handle()) << std::endl;
    std::cout << *strategy->get_priority_settings(listener2->sub_->get_subscription_handle()) << std::endl;
    executor->spin();

    rclcpp::shutdown();

    // Get all collected metrics and write them to a file
    collector.write_logs_to_file("uas", "existing_system_monitor");
    
    return 0;
}